
module Main (main) where

import System.Exit (exitWith, ExitCode(ExitFailure))
import Maybe (fromMaybe)
import List (elemIndex, sort, nub)

import qualified Data.List.LCS.Simple as Simple (lcs)
import qualified Data.List.LCS.OTHER as Other (lcs)

-- Maximum size of test case (== length (xs ++ ys)) to use
max_size :: Int
max_size = 7

-- Maximum size of test case to test test-case generation with
tg_max_size :: Int
tg_max_size = 5

-- Run the tests
main :: IO ()
main = do putStrLn ("Running TG with a max_size of " ++ show tg_max_size)
          case test_generation_test tg_max_size of
              [] -> putStrLn "Passed"
              fs -> do putStrLn "Failed"
                       mapM_ (putStrLn . show_TG_failure) fs
                       putStrLn (show (length fs) ++ " failures")
                       exitWith (ExitFailure 1)
          putStrLn ("Running with a max_size of " ++ show max_size)
          if null test_failures
            then putStrLn "All good!"
            else do putStrLn "Failed"
                    mapM_ (putStrLn . show_failure) test_failures
                    putStrLn (show (length test_failures) ++ " failures")
                    exitWith (ExitFailure 1)

-- Show testcase generation failures
show_TG_failure :: Either (String, String) (String, String) -> String
show_TG_failure (Left (xs, ys)) = show (xs, ys) ++ " generated by spec only"
show_TG_failure (Right (xs, ys)) = show (xs, ys) ++ " generated by impl only"

-- Show test failures
show_failure :: (String, String) -> String
show_failure (xs, ys) = "Error computing LCS of " ++ show xs
                                       ++ " and " ++ show ys

-- The list of test failures
test_failures :: [(String, String)]
test_failures = [ p | p@(xs, ys) <- test_cases max_size, not $ do_test xs ys ]

-- Returns True iff the implementation gives a valid answer for the inputs
do_test :: String -> String -> Bool
do_test xs ys = let zs_simple = Simple.lcs xs ys
                    zs_other = Other.lcs xs ys
                in zs_other `is_subseq_of` xs &&
                   zs_other `is_subseq_of` ys &&
                   length zs_simple == length zs_other

-- Returns True iff the first argument is a subsequence of the second
is_subseq_of :: String -> String -> Bool
is_subseq_of (x:xs) (y:ys) | x == y = is_subseq_of xs ys
                           | otherwise = is_subseq_of (x:xs) ys
is_subseq_of [] _ = True
is_subseq_of _ [] = False

-- Test case generation specification

test_cases_spec :: Int -> [(String, String)]
test_cases_spec n = concatMap do_splits $ filter valid $ strings_from 0 [""]
    where -- The alphabet we use
          chars = take n ['a'..]
          -- Make all strings from the alphabet of the desired length
          strings_from i xs
           | i == n    = xs
           | otherwise = xs ++ strings_from (i+1)
                                            [ c:cs | cs <- xs, c <- chars ]
          -- A test case is valid only if it starts at 'a' and misses
          -- no letters out
          valid xs = is 'a' $ nub $ sort xs
          is c (x:xs) = c == x && is (succ c) xs
          is _ [] = True
          -- Wrapper for splits where the length is not supplied
          do_splits xs = splits (length xs) xs

-- Test case generation

test_cases :: Int -> [(String, String)]
test_cases 0 = [("", "")]
test_cases n = ("", ""):gen_cases n 1 "a"

gen_cases :: Int -> Int -> String -> [(String, String)]
gen_cases n_max n s | n_max == n = splits n s
gen_cases n_max n s = splits n s
                   ++ concatMap (gen_cases n_max (n+1)) (add_char n s)

splits :: Int -> String -> [(String, String)]
splits n s = [ splitAt i s | i <- [0..n] ]

add_char :: Int -> String -> [String]
add_char n s = add_first_a n s ++ add_a n (map succ s)

add_a :: Int -> String -> [String]
add_a n s = map (\(xs, ys) -> xs ++ "a" ++ ys) (splits n s)

add_first_a :: Int -> String -> [String]
add_first_a n s = add_a (fromMaybe n (elemIndex 'a' s)) s

-- Test case generation test (mmm, meta)

test_generation_test :: Int -> [Either (String, String) (String, String)]
test_generation_test n = f (sort (test_cases_spec n)) (sort (test_cases n))
    where f [] [] = []
          f xs [] = map Left xs
          f [] ys = map Right ys
          f (x:xs) (y:ys) = case compare xs ys of
                                EQ -> f xs ys
                                LT -> Left x:f xs (y:ys)
                                GT -> Right y:f (x:xs) ys

